span-viewer {
  min-width: 0;
  display: flex;
  flex: 1;
  align-self: stretch;
  flex-direction: column;
  gap: 20px;

  h3 {
    align-self: stretch;
  }

  & > section {
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    align-self: stretch;

    pre {
      align-self: stretch;
      max-width: 100%;
    }


    &.grow {
      flex: 1;
    }
  }
}
